@@ -1,5 +1,5 @@
 /***************************************************************************************************
- * Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * Copyright (c) 2017 - 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: BSD-3-Clause
  *
  * Redistribution and use in source and binary forms, with or without
@@ -28,6 +28,16 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  **************************************************************************************************/
+/*
+  Note:  CUTLASS 3x increases the host compiler requirements to C++17. However, certain
+         existing integrations of CUTLASS require C++11 host compilers.
+
+         Until this requirement can be lifted, certain headers with this annotation are required
+         to be remain consistent with C++11 syntax.
+
+         C++11 compatibility is enforced by this unit test: `cutlass_test_unit_core_cpp11`.
+*/
+
 #pragma once
 
 #include <cuComplex.h>
@@ -42,11 +52,9 @@
 
 #include "cutlass/cutlass.h"
 #include "cutlass/functional.h"
-#include "cutlass/half.h"
 #include "cutlass/real.h"
 
-#include "cutlass/bfloat16.h"
-#include "cutlass/tfloat32.h"
+#include "cutlass/numeric_types.h"
 
 #include "cutlass/fast_math.h"
 
@@ -199,7 +207,7 @@ class complex
   template <typename OtherT>
   CUTLASS_DEVICE void red(complex<OtherT> *ptr) const {
     static_assert(platform::is_same<T, OtherT>::value, "Component type must match");
-    cutlass::red<T> reduce;
+    cutlass::atomic_add<T> reduce;
     reduce(&ptr->_real, _real);
     reduce(&ptr->_imag, _imag);
   }
@@ -209,7 +217,7 @@ class complex
     static_assert(platform::is_same<T, half_t>::value, "Component type must match");
     half2 *h2_ptr = reinterpret_cast<half2*>(ptr);
     half2 h2_data = reinterpret_cast<half2&>(*this);
-    cutlass::red<half2> reduce;
+    cutlass::atomic_add<half2> reduce;
     reduce(h2_ptr, h2_data);
   }
 
@@ -300,6 +308,13 @@ class complex
   CUTLASS_HOST_DEVICE
   T &imag() { return _imag; }
 
+  /// Set the real part of the complex number
+  CUTLASS_HOST_DEVICE
+  void real(T real) { _real = real; }
+
+  /// Set the imaginary part of the complex number
+  CUTLASS_HOST_DEVICE
+  void imag(T imag) { _imag = imag; }
 
   #if !defined(__CUDACC_RTC__)
   /// Converts to cuFloatComplex
@@ -431,34 +446,72 @@ CUTLASS_HOST_DEVICE R norm_accumulate(T const &x, R const & accumulator) {
 /// Norm accumulate specialized for complex types
 template <typename T, typename R>
 CUTLASS_HOST_DEVICE R norm_accumulate(complex<T> const &z, R const &accumulator) {
-  return accumulator + static_cast<R>(real(z)) * static_cast<R>(real(z)) + 
+  return accumulator + static_cast<R>(real(z)) * static_cast<R>(real(z)) +
     static_cast<R>(imag(z)) * static_cast<R>(imag(z));
 }
 
-/// Returns the complex conjugate
 CUTLASS_HOST_DEVICE float conj(float const &z) {
   return z;
 }
 
-/// Returns the complex conjugate
 CUTLASS_HOST_DEVICE double conj(double const &z) {
   return z;
 }
 
+CUTLASS_HOST_DEVICE half_t conj(half_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE int32_t conj(int32_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE uint32_t conj(uint32_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE int64_t conj(int64_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE uint64_t conj(uint64_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE int4b_t conj(int4b_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE uint4b_t conj(uint4b_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE bfloat16_t conj(bfloat16_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE uint1b_t conj(uint1b_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE tfloat32_t conj(tfloat32_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE float_e4m3_t conj(float_e4m3_t const& z) {
+  return z;
+}
+
+CUTLASS_HOST_DEVICE float_e5m2_t conj(float_e5m2_t const& z) {
+  return z;
+}
+
+
 /// Returns the complex conjugate
 template <typename T>
 CUTLASS_HOST_DEVICE complex<T> conj(complex<T> const &z) {
   return complex<T>(real(z), -imag(z));
 }
-/// Indentity transform for non-complex types
-template <typename T>
-CUTLASS_HOST_DEVICE T conj(T const &z) {
-    static_assert( !platform::is_same<T, cuComplex>::value &&
-                   !platform::is_same<T, cuDoubleComplex>::value &&
-                   !platform::is_same<T, cutlass::complex<double>>::value &&
-                   !platform::is_same<T, cutlass::complex<float>>::value, "May not be a complex data type");
-  return z;
-}
 
 /// Projects the complex number z onto the Riemann sphere
 template <typename T>
@@ -511,11 +564,10 @@ CUTLASS_HOST_DEVICE complex<T> sin(complex<T> const &z) {
   return (exp(-z) - exp(z)) * complex<T>(T(0), T(1) / T(2));
 }
 
-/// Comparison 
+/// Comparison
 template <typename T>
 CUTLASS_HOST_DEVICE bool operator<(complex<T> const &lhs, complex<T> const &rhs) {
-  //TODO
-  return true; 
+  return true;
 }
 
 //////////////////////////////////////////////////////////////////////////////////////////////////
@@ -679,7 +731,7 @@ struct magnitude_squared_difference<complex<T>, Output> {
 
 /// Reduces value into the data pointed to by ptr (complex<T> specialization)
 template <typename T>
-struct red<complex<T>> {
+struct atomic_add<complex<T>> {
   CUTLASS_DEVICE
   void operator()(complex<T> *ptr, const complex<T> &data)
   {
