@@ -1,5 +1,5 @@
 /***************************************************************************************************
- * Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * Copyright (c) 2017 - 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: BSD-3-Clause
  *
  * Redistribution and use in source and binary forms, with or without
@@ -140,6 +140,8 @@ public:
     typename LayoutC::Stride::Index ldc;
     typename LayoutC::Stride::Index ldd;
 
+    bool allow_early_exit;
+
     //
     // Methods
     //
@@ -147,7 +149,8 @@ public:
     Arguments(): 
       mode(GemmUniversalMode::kGemm), 
       batch_count(1), 
-      ptr_A(nullptr), ptr_B(nullptr), ptr_C(nullptr), ptr_D(nullptr) { }
+      ptr_A(nullptr), ptr_B(nullptr), ptr_C(nullptr), ptr_D(nullptr),
+      allow_early_exit(false) { }
 
     /// constructs an arguments structure
     Arguments(
@@ -166,7 +169,8 @@ public:
       typename LayoutA::Stride::Index lda,
       typename LayoutB::Stride::Index ldb,
       typename LayoutC::Stride::Index ldc,
-      typename LayoutC::Stride::Index ldd
+      typename LayoutC::Stride::Index ldd,
+      bool allow_early_exit = false
     ):
       mode(mode), 
       problem_size(problem_size), 
@@ -174,7 +178,8 @@ public:
       epilogue(epilogue), 
       ptr_A(ptr_A), ptr_B(ptr_B), ptr_C(ptr_C), ptr_D(ptr_D), 
       batch_stride_A(batch_stride_A), batch_stride_C(batch_stride_C), batch_stride_D(batch_stride_D), 
-      lda(lda), ldb(ldb), ldc(ldc), ldd(ldd) {
+      lda(lda), ldb(ldb), ldc(ldc), ldd(ldd),
+      allow_early_exit(allow_early_exit) {
 
       }
 
@@ -231,6 +236,8 @@ public:
 
     int *semaphore;
 
+    bool allow_early_exit;
+
     //
     // Methods
     //
@@ -255,7 +262,8 @@ public:
       batch_stride_B(0),
       batch_stride_C(0),
       batch_stride_D(0),
-      semaphore(nullptr) { }
+      semaphore(nullptr),
+      allow_early_exit(false) { }
 
     CUTLASS_HOST_DEVICE
     Params(
@@ -285,7 +293,8 @@ public:
       batch_stride_B(args.batch_stride_B),
       batch_stride_C(args.batch_stride_C),
       batch_stride_D(args.batch_stride_D),
-      semaphore(static_cast<int *>(workspace)) {
+      semaphore(static_cast<int *>(workspace)),
+      allow_early_exit(args.allow_early_exit) {
     }
 
     CUTLASS_HOST_DEVICE
@@ -347,6 +356,12 @@ public:
   CUTLASS_DEVICE
   void operator()(Params const &params, SharedStorage &shared_storage) {
 
+    // Early exit following LAPACK's definition
+    if (params.allow_early_exit &&
+        (params.output_op.alpha == ElementC(0)) && (params.output_op.beta == ElementC(1))) {
+      return;
+    }
+
     // Compute threadblock location
     ThreadblockSwizzle threadblock_swizzle;
 
@@ -450,7 +465,7 @@ public:
 
     // Broadcast the warp_id computed by lane 0 to ensure dependent code
     // is compiled as warp-uniform.
-    int warp_idx = canonical_warp_idx();
+    int warp_idx = canonical_warp_idx_sync();
 
     int lane_idx = threadIdx.x % 32;
 
