@@ -1,6 +1,6 @@
 #################################################################################################
 #
-# Copyright (c) 2017 - 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+# Copyright (c) 2017 - 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 # SPDX-License-Identifier: BSD-3-Clause
 #
 # Redistribution and use in source and binary forms, with or without
@@ -34,21 +34,42 @@ import ctypes
 import json
 import os
 import sqlite3
+import subprocess
 import tempfile
 
 from cuda import cuda, nvrtc
-import cutlass_bindings
+from cutlass_library import SubstituteTemplate
 
-from cutlass import CACHE_FILE, CUDA_INSTALL_PATH, CUTLASS_PATH
+import cutlass
+from cutlass import CACHE_FILE, CUTLASS_PATH, cuda_install_path, logger
 from cutlass.backend.gemm_operation import GemmOperationUniversal
 from cutlass.backend.library import ApiVersion
 from cutlass.backend.utils.device import device_cc
-from cutlass.backend.utils.software import SubstituteTemplate
 
 IncludeTemplate = r"""#include "${include}"
 """
 
 
+def compile_with_nvcc(cmd, source, error_file):
+    succeed = True
+    try:
+        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
+    except subprocess.CalledProcessError as e:
+        error_message = e.output.decode()
+        with open(error_file, "w") as error_out:
+            error_log = "Compilation error for the following kernel: \n"
+            error_log += source
+            error_log += "\nError Message:\n"
+            error_log += error_message
+            error_out.write(error_log)
+        succeed = False
+    if not succeed:
+        # Print the error log to stdout if log level is set to warning or higher
+        # verbosity. Otherwise, simply point to the error log file.
+        logger.warning(error_log)
+        raise Exception(f"Invalid Kernel. See '{error_file}' for details.")
+
+
 class CompilationOptions:
     """
     Compilation options.
@@ -61,20 +82,19 @@ class CompilationOptions:
         self.arch = arch
 
     def get_str(self):
-        options = ""
-
+        opts = []
         for flag in self.flags:
-            options += " " + flag
+            opts.append(flag)
 
         for incl in self.include_paths:
-            options += " --include-path=%s" % incl
+            opts.append(f"--include-path={incl}")
 
-        arch_flag = " -arch=sm_%d" % self.arch
+        arch_flag = f"-arch=sm_{self.arch}"
         if self.arch == 90:
             arch_flag += "a"
-        options += arch_flag
+        opts.append(arch_flag)
 
-        return options
+        return " ".join(opts)
 
     def get(self):
         options = []
@@ -83,9 +103,9 @@ class CompilationOptions:
             options.append(bytes(str.encode(flag)))
 
         for incl in self.include_paths:
-            options.append(bytes(str.encode("--include-path=%s" % incl)))
+            options.append(bytes(str.encode(f" --include-path={incl}")))
 
-        arch_flag = " -arch=sm_%d" % self.arch
+        arch_flag = f" -arch=sm_{self.arch}"
         if self.arch == 90:
             arch_flag += "a"
 
@@ -129,20 +149,24 @@ class ArtifactManager:
         connection.commit()
         cursor.close()
 
-        self.nvcc()
-        self.compiled_cache_device = cutlass_bindings.CompileCache()
-        self.compiled_cache_host = cutlass_bindings.CompileCache()
-
-    def nvrtc(self):
-        self.backend = "nvrtc"
-        self.default_compile_options = ["-std=c++17", "-default-device"]
-    def nvcc(self):
-        self.backend = "nvcc"
-        self.default_compile_options = [
+        self._nvrtc_compile_options = ["-std=c++17", "-default-device"]
+        self._nvcc_compile_options = [
             "-std=c++17",
             "--expt-relaxed-constexpr",
             "-Xcudafe --diag_suppress=esa_on_defaulted_function_ignored",
         ]
+        self.nvcc()
+        self.compiled_cache_device = {}
+        self.compiled_cache_host = {}
+
+    def nvrtc(self):
+        self.backend = "nvrtc"
+        self.default_compile_options = self._nvrtc_compile_options
+
+    def nvcc(self):
+        self.backend = "nvcc"
+        self.default_compile_options = self._nvcc_compile_options
+
     def insert_operation(self, op_key, cubin, hostfile, op_name, op_attrs):
         connection = sqlite3.connect(CACHE_FILE)
         cursor = connection.cursor()
@@ -172,7 +196,7 @@ class ArtifactManager:
                 raise RuntimeError("Cuda Error: {}".format(err))
 
             err, kernel = cuda.cuModuleGetFunction(module, bytes(str.encode(operation_name)))
-            self.compiled_cache_device.insert(key, kernel)
+            self.compiled_cache_device[key] = kernel
 
             compiled_host_fns = {}
             host_lib = CDLLBin(host_binary)
@@ -197,10 +221,10 @@ class ArtifactManager:
 
                     compiled_host_fns[attr] = func
 
-            self.compiled_cache_host.insert(key, compiled_host_fns)
+            self.compiled_cache_host[key] = compiled_host_fns
         return True
 
-    def emit_compile_(self, operation_list, compilation_options):
+    def emit_compile_(self, operation_list, compilation_options, host_compilation_options):
         """
         Compile a list of kernels and store them into database
         """
@@ -221,11 +245,10 @@ class ArtifactManager:
             )
 
         for incl in includes_host:
-            if "/device/" not in incl:
-                source_buffer_host += SubstituteTemplate(
-                    IncludeTemplate,
-                    {"include": incl},
-                )
+            source_buffer_host += SubstituteTemplate(
+                IncludeTemplate,
+                {"include": incl},
+            )
 
         # 2. Operations
         for operation in operation_list:
@@ -293,59 +316,59 @@ class ArtifactManager:
             # compile with nvcc
             cmd_template = "${cuda_install_path}/bin/nvcc ${options} -cubin ${srcfile} -o ${tarfile}"
             values = {
-                "cuda_install_path": CUDA_INSTALL_PATH,
+                "cuda_install_path": cuda_install_path(),
                 "options": compilation_options.get_str(),
                 "srcfile": temp_cu.name,
                 "tarfile": temp_cubin.name,
             }
             cmd = SubstituteTemplate(cmd_template, values)
-            os.system(cmd)
+            compile_with_nvcc(cmd.split(" "), source_buffer_device, "./cutlass_python_compilation_device_error.txt")
 
             # load the cubin image
             with open(temp_cubin.name, "rb") as file:
                 cubin_image = file.read()
 
-        # Set up the host-side library code
-        cmd_template = (
-            "echo '%s'|${cuda_install_path}/bin/nvcc -x cu -Xcompiler=\"-fpermissive -w -fPIC\" ${options}"
-            % source_buffer_host
-        )
-        cmd = SubstituteTemplate(
-            cmd_template,
-            {
-                "cuda_install_path": CUDA_INSTALL_PATH,
-                "options": compilation_options.get_str(),
-            },
-        )
-
         tempfile.tempdir = "./"
-        temp = tempfile.NamedTemporaryFile(
+        temp_src = tempfile.NamedTemporaryFile(
+            prefix="host_src", suffix=".cu", delete=True)
+
+        # Write the host source
+        with open(temp_src.name, "w") as outfile:
+            outfile.write(source_buffer_host)
+
+        temp_dst = tempfile.NamedTemporaryFile(
             prefix="host_func", suffix=".so", delete=True)
 
-        cmd += " - -shared -o %s -lcudart -lcuda" % temp.name
-        os.system(cmd)
-        host_lib = ctypes.CDLL(temp.name)
+        # Set up host compilation arguments
+        cmd = []
+        cmd.append(f"{cuda_install_path()}/bin/nvcc")
+        cmd.extend(["-x", "cu", "-Xcompiler=-fpermissive", "-Xcompiler=-w", "-Xcompiler=-fPIC"])
+        cmd.extend(host_compilation_options.get_str().split(" "))
+        cmd.extend(["-shared", "-o", temp_dst.name, temp_src.name, "-lcudart", "-lcuda"])
 
-        return cubin_image, host_lib, temp
+        # Comile and load the library
+        compile_with_nvcc( cmd, source_buffer_host, error_file="./cutlass_python_compilation_host_error.txt")
+        host_lib = ctypes.CDLL(temp_dst.name)
 
-    def add_module(self, operations, compile_options=None):
+        return cubin_image, host_lib, temp_dst
+
+    def add_module(self, operations, compile_options=None, bypass_cache=False):
         """
         Insert a new compiled device module
         """
-        if compile_options is None:
-            include_paths = [
-                CUDA_INSTALL_PATH + "/include",
-                CUTLASS_PATH + "/include",
-                CUTLASS_PATH + "/tools/util/include",
-                CUTLASS_PATH + "/python/cutlass/cpp/include",
-            ]
+        include_paths = [
+            cuda_install_path() + "/include",
+            CUTLASS_PATH + "/include",
+            CUTLASS_PATH + "/tools/util/include",
+            CUTLASS_PATH + "/python/cutlass/cpp/include",
+        ]
 
-            if device_cc() is not None:
-                arch = device_cc()
-            else:
-                # Find the maximum arch tag among the provided operations and compile for that target.
-                # Since we are compiling to .cubin files, only one architecture may be specified.
-                arch = max([op.arch for op in operations])
+        cutlass.initialize_cuda_context()
+        arch = device_cc()
+
+        host_compile_options = CompilationOptions(
+            self._nvcc_compile_options, arch, include_paths)
+        if compile_options is None:
             compile_options = CompilationOptions(
                 self.default_compile_options, arch, include_paths)
         # save the cubin
@@ -355,16 +378,16 @@ class ArtifactManager:
             # step 1: get kernel string as key
             key = operation.rt_module.emit() + operation.procedural_name() + self.backend
             # step 1: check if the operation is in cache
-            compiled_kernel = self.compiled_cache_device.at(key)
+            compiled_kernel = self.compiled_cache_device.get(key)
 
-            if compiled_kernel is None:
+            if compiled_kernel is None and not bypass_cache:
                 hit = self.load_operation(key, getattr( operation.rt_module, "extra_funcs", {}))
                 if hit:
-                    compiled_kernel = self.compiled_cache_device.at(key)
+                    compiled_kernel = self.compiled_cache_device.get(key)
                     assert compiled_kernel is not None
             if compiled_kernel is not None:
                 operation.rt_module.kernel = compiled_kernel
-                compiled_host_fns = self.compiled_cache_host.at(key)
+                compiled_host_fns = self.compiled_cache_host.get(key)
                 assert compiled_host_fns is not None
                 for key in compiled_host_fns.keys():
                     setattr(operation.rt_module, key, compiled_host_fns[key])
@@ -375,7 +398,7 @@ class ArtifactManager:
 
         if len(operation_list) > 0:
             cubin_image, host_lib, host_file = self.emit_compile_(
-                operation_list, compile_options)
+                operation_list, compile_options, host_compile_options)
 
             err, module = cuda.cuModuleLoadData(cubin_image)
             if err != cuda.CUresult.CUDA_SUCCESS:
@@ -390,7 +413,7 @@ class ArtifactManager:
                     bytes(str.encode(operation.name()))
                 )
                 operation_name.append(operation.name())
-                self.compiled_cache_device.insert(key, operation.kernel)
+                self.compiled_cache_device[key] = operation.kernel
                 # get host functions
                 compiled_host_fns = {}
                 op_attr = []
@@ -429,7 +452,7 @@ class ArtifactManager:
                         op_attr.append(suffix)
 
                 operation_attr.append(op_attr)
-                self.compiled_cache_host.insert(key, compiled_host_fns)
+                self.compiled_cache_host[key] = compiled_host_fns
 
             for (key, operation_name, operation_attr,) in zip(operation_key, operation_name, operation_attr):
                 self.insert_operation(
