@@ -1,5 +1,5 @@
 /***************************************************************************************************
- * Copyright (c) 2017 - 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * Copyright (c) 2017 - 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: BSD-3-Clause
  *
  * Redistribution and use in source and binary forms, with or without
@@ -38,6 +38,7 @@
 #include <utility>
 #include <cstdlib>
 #include <cmath>
+#include <random>
 
 // Cutlass includes
 #include "cutlass/cutlass.h"
@@ -92,6 +93,25 @@ struct TensorFillFunc {
   }
 };
 
+/// Returns a pair of values of the Gaussian distribution generated by the Box Muller method 
+struct BoxMullerFunc {
+
+  BoxMullerFunc() {}
+
+  void operator()(
+    double* rnd,                     ///< Size-2 vector to be filled with random values
+    double  mean = 0,                ///< Mean of the Gaussian distribution
+    double  stddev = 1,              ///< Standard deviation of the Gaussian distribution
+    double  pi = std::acos(-1)) const {
+
+    double u1 = double(std::rand()) / double(RAND_MAX);
+    double u2 = double(std::rand()) / double(RAND_MAX);
+    rnd[0] = std::sqrt(-2 * std::log(u1)) * std::cos(2 * pi * u2);
+    rnd[1] = std::sqrt(-2 * std::log(u1)) * std::sin(2 * pi * u2);
+    rnd[0] = mean + stddev * rnd[0];
+    rnd[1] = mean + stddev * rnd[1];
+  }
+};
 } // namespace detail
 
 ///////////////////////////////////////////////////////////////////////////////////////////////////
@@ -137,6 +157,7 @@ struct RandomGaussianFunc {
   double stddev;
   int int_scale;
   double pi;
+  double pnz;
 
   //
   // Methods
@@ -145,9 +166,10 @@ struct RandomGaussianFunc {
     uint64_t seed_ = 0, 
     double mean_ = 0, 
     double stddev_ = 1,
-    int int_scale_ = -1
+    int int_scale_ = -1,
+    double pnz_ = 100.0
   ):
-    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)) {
+    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)), pnz(pnz_) {
       std::srand((unsigned)seed);
   }
 
@@ -165,12 +187,24 @@ struct RandomGaussianFunc {
     // Scale and convert final result
     Element result;
 
-    if (int_scale >= 0) {
-      rnd = double(int64_t(rnd * double(1 << int_scale))) / double(1 << int_scale);
-      result = static_cast<Element>(rnd);
+    // Sample from the Bernoulli distribution, and use the result to sample from the Gaussian
+    std::random_device rnd_device;
+    std::mt19937 bernoulli_rnd(rnd_device());
+    std::bernoulli_distribution bernoulli_dist(pnz / 100);
+    bool bernoulli_result = bernoulli_dist(bernoulli_rnd);
+
+    // Sample from the Gaussian distribution for a nonzero element
+    if (bernoulli_result) {
+      if (int_scale >= 0) {
+        rnd = double(int64_t(rnd * double(1 << int_scale))) / double(1 << int_scale);
+        result = static_cast<Element>(rnd);
+      }
+      else {
+        result = static_cast<Element>(rnd);
+      }
     }
     else {
-      result = static_cast<Element>(rnd);
+      result = static_cast<Element>(0);
     }
 
     return result;
@@ -186,6 +220,7 @@ struct RandomGaussianFunc<complex<Element> > {
   double stddev;
   int int_scale;
   double pi;
+  double pnz;
 
   //
   // Methods
@@ -194,9 +229,10 @@ struct RandomGaussianFunc<complex<Element> > {
     uint64_t seed_ = 0, 
     double mean_ = 0, 
     double stddev_ = 1,
-    int int_scale_ = -1
+    int int_scale_ = -1,
+    double pnz_ = 100.0
   ):
-    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)) {
+    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)), pnz(pnz_) {
       std::srand((unsigned)seed);
   }
 
@@ -205,23 +241,33 @@ struct RandomGaussianFunc<complex<Element> > {
 
     Element reals[2];
 
-    for (int i = 0; i < 2; ++i) {
-      // Box-Muller transform to generate random numbers with Normal distribution
-      double u1 = double(std::rand()) / double(RAND_MAX);
-      double u2 = double(std::rand()) / double(RAND_MAX);
+    double rnd[2];
+    detail::BoxMullerFunc func;
+    func(rnd, mean, stddev, pi);
 
-      // Compute Gaussian random value
-      double rnd = std::sqrt(-2 * std::log(u1)) * std::cos(2 * pi * u2);
-      rnd = mean + stddev * rnd;
+    // Sample from the Bernoulli distribution, and use the result to sample from the Gaussian
+    std::random_device rnd_device;
+    std::mt19937 bernoulli_rnd(rnd_device());
+    std::bernoulli_distribution bernoulli_dist(pnz / 100);
+    bool bernoulli_result = bernoulli_dist(bernoulli_rnd);
 
+    // Sample from the Gaussian distribution for a nonzero element
+    if (bernoulli_result) {
       if (int_scale >= 0) {
-        rnd = double(int(rnd * double(1 << int_scale)));
-        reals[i] = from_real<Element>(rnd / double(1 << int_scale));
+        rnd[0] = double(int(rnd[0] * double(1 << int_scale)));
+        rnd[1] = double(int(rnd[1] * double(1 << int_scale)));
+        reals[0] = from_real<Element>(rnd[0] / double(1 << int_scale));
+        reals[1] = from_real<Element>(rnd[1] / double(1 << int_scale));
       }
       else {
-        reals[i] = from_real<Element>(rnd);
+        reals[0] = from_real<Element>(rnd[0]);
+        reals[1] = from_real<Element>(rnd[1]);
       }
     }
+    else {
+      reals[0] = from_real<Element>(0);
+      reals[1] = from_real<Element>(0);
+    }
 
     return complex<Element>(reals[0], reals[1]);
   }
@@ -236,6 +282,7 @@ struct RandomGaussianFunc<Quaternion<Element> > {
   double stddev;
   int int_scale;
   double pi;
+  double pnz;
 
   //
   // Methods
@@ -244,9 +291,10 @@ struct RandomGaussianFunc<Quaternion<Element> > {
     uint64_t seed_ = 0,
     double mean_ = 0,
     double stddev_ = 1,
-    int int_scale_ = -1
+    int int_scale_ = -1,
+    double pnz_ = 100.0
   ):
-    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)) {
+    seed(seed_), mean(mean_), stddev(stddev_), int_scale(int_scale_), pi(std::acos(-1)), pnz(pnz_) {
       std::srand((unsigned)seed);
   }
 
@@ -255,23 +303,44 @@ struct RandomGaussianFunc<Quaternion<Element> > {
 
     Element reals[4];
 
-    for (int i = 0; i < 4; ++i) {
-      // Box-Muller transform to generate random numbers with Normal distribution
-      double u1 = double(std::rand()) / double(RAND_MAX);
-      double u2 = double(std::rand()) / double(RAND_MAX);
+    double rnd1[2];
+    double rnd2[2];
+    detail::BoxMullerFunc func;
+    func(rnd1, mean, stddev, pi);
+    func(rnd2, mean, stddev, pi);
 
-      // Compute Gaussian random value
-      double rnd = std::sqrt(-2 * std::log(u1)) * std::cos(2 * pi * u2);
-      rnd = mean + stddev * rnd;
+    // Sample from the Bernoulli distribution, and use the result to sample from the Gaussian
+    std::random_device rnd_device;
+    std::mt19937 bernoulli_rnd(rnd_device());
+    std::bernoulli_distribution bernoulli_dist(pnz / 100);
+    bool bernoulli_result = bernoulli_dist(bernoulli_rnd);
 
+    // Sample from the Gaussian distribution for a nonzero element
+    if (bernoulli_result) {
       if (int_scale >= 0) {
-        rnd = double(int(rnd * double(1 << int_scale)));
-        reals[i] = from_real<Element>(rnd / double(1 << int_scale));
+        rnd1[0] = double(int(rnd1[0] * double(1 << int_scale)));
+        rnd1[1] = double(int(rnd1[1] * double(1 << int_scale)));
+        rnd2[0] = double(int(rnd2[0] * double(1 << int_scale)));
+        rnd2[1] = double(int(rnd2[1] * double(1 << int_scale)));
+
+        reals[0] = from_real<Element>(rnd1[0] / double(1 << int_scale));
+        reals[1] = from_real<Element>(rnd1[1] / double(1 << int_scale));
+        reals[2] = from_real<Element>(rnd2[0] / double(1 << int_scale));
+        reals[3] = from_real<Element>(rnd2[1] / double(1 << int_scale));
       }
       else {
-        reals[i] = from_real<Element>(rnd);
+        reals[0] = from_real<Element>(rnd1[0]);
+        reals[1] = from_real<Element>(rnd1[1]);
+        reals[2] = from_real<Element>(rnd2[0]);
+        reals[3] = from_real<Element>(rnd2[1]);
       }
     }
+    else {
+      reals[0] = from_real<Element>(0);
+      reals[1] = from_real<Element>(0);
+      reals[2] = from_real<Element>(0);
+      reals[3] = from_real<Element>(0);
+    }
 
     return Quaternion<Element>(reals[0], reals[1], reals[2], reals[3]);
   }
@@ -311,7 +380,7 @@ struct TensorFillGaussianFunc {
   }
 };
 
-/// Computes a random Gaussian distribution
+/// Computes a random Gaussian distribution for a rank-2 tensor
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -369,11 +438,11 @@ void TensorFillRandomGaussian(
   uint64_t seed,                          ///< seed for RNG
   double mean = 0,                        ///< Gaussian distribution's mean
   double stddev = 1,                      ///< Gaussian distribution's standard deviation
-  int bits = -1) {                        ///< If non-negative, specifies number of fractional bits that 
-                                          ///  are not truncated to zero. Permits reducing precision of
+  int bits = -1,                          ///< If non-negative, specifies number of fractional bits that 
+  double pnz = 100.0) {                   ///  are not truncated to zero. Permits reducing precision of
                                           ///  data.
   
-  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits);
+  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits, pnz);
 
   detail::TensorFillGaussianFunc<Element, Layout> func(
     dst,
@@ -391,20 +460,20 @@ template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
 void TensorFillRandomGaussian(
-  TensorViewPlanarComplex<Element, Layout> dst,        ///< destination tensor
-  uint64_t seed,                                       ///< seed for RNG
-  double mean = 0,                                     ///< Gaussian distribution's mean
-  double stddev = 1,                                   ///< Gaussian distribution's standard deviation
-  int bits = -1) {                                     ///< If non-negative, specifies number of fractional bits that 
-                                                       ///  are not truncated to zero. Permits reducing precision of
-                                                       ///  data.
+  TensorViewPlanarComplex<Element, Layout> dst,         ///< destination tensor
+  uint64_t seed,                                        ///< seed for RNG
+  double mean = 0,                                      ///< Gaussian distribution's mean
+  double stddev = 1,                                    ///< Gaussian distribution's standard deviation
+  int bits = -1,                                        ///< If non-negative, specifies number of fractional bits that 
+  double pnz = 100.0) {                                 ///  are not truncated to zero. Permits reducing precision of
+                                                        ///  data.
   
-  TensorFillRandomGaussian(dst.view_real(), seed, mean, stddev, bits);
-  TensorFillRandomGaussian(dst.view_imag(), ~seed, mean, stddev, bits);
+  TensorFillRandomGaussian(dst.view_real(), seed, mean, stddev, bits, pnz);
+  TensorFillRandomGaussian(dst.view_imag(), ~seed, mean, stddev, bits, pnz);
 }
 
 ///////////////////////////////////////////////////////////////////////////////////////////////////
-/// Fills a tensor with random values with a Gaussian distribution.
+/// Fills the upper or lower part of a symmetric rank-2 tensor with random values of a Gaussian distribution.
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -414,11 +483,11 @@ void TensorFillSymmetricRandomGaussian(
   cutlass::FillMode fill_mode,            ///< FillMode for symmetric matrices
   double mean = 0,                        ///< Gaussian distribution's mean
   double stddev = 1,                      ///< Gaussian distribution's standard deviation
-  int bits = -1) {                        ///< If non-negative, specifies number of fractional bits that 
-                                          ///  are not truncated to zero. Permits reducing precision of
+  int bits = -1,                          ///< If non-negative, specifies number of fractional bits that 
+  double pnz = 100.0) {                   ///  are not truncated to zero. Permits reducing precision of
                                           ///  data.
 
-  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits);
+  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits, pnz);
 
   detail::TensorFillSymmetricGaussianFunc<Element, Layout> func(
     dst,
@@ -434,7 +503,7 @@ void TensorFillSymmetricRandomGaussian(
 
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
-/// Fills a tensor with random values with a Gaussian distribution.
+/// Fills a tensor with random values of a Gaussian distribution.
 template <
   typename Element                        ///< Element type
 >
@@ -444,12 +513,12 @@ void BlockFillRandomGaussian(
   uint64_t seed,                          ///< seed for RNG
   double mean = 0,                        ///< Gaussian distribution's mean
   double stddev = 1,                      ///< Gaussian distribution's standard deviation
-  int bits = -1) {                        ///< If non-negative, specifies number of fractional bits that 
-                                          ///  are not truncated to zero. Permits reducing precision of
+  int bits = -1,                          ///< If non-negative, specifies number of fractional bits that 
+  double pnz = 100.0) {                   ///  are not truncated to zero. Permits reducing precision of
                                           ///  data.
   
 
-  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits);
+  detail::RandomGaussianFunc<Element> random_func(seed, mean, stddev, bits, pnz);
 
   for (size_t i = 0; i < capacity; ++i) {
     ReferenceFactory<Element>::get(ptr, i) = random_func();
@@ -497,7 +566,6 @@ struct RandomUniformFunc {
     // Random values are cast to integer after scaling by a power of two to facilitate error
     // testing
     Element result;
-    
     if (int_scale >= 0) {
       rnd = double(int64_t(rnd * double(1 << int_scale))) / double(1 << int_scale);
       result = static_cast<Element>(Real(rnd));
@@ -614,7 +682,7 @@ struct RandomUniformFunc<Quaternion<Element> > {
   }
 };
 
-/// Computes a random Gaussian distribution
+/// Computes a random uniform distribution
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -633,7 +701,7 @@ struct TensorFillRandomUniformFunc {
   // Methods
   //
 
-  /// Construction of Gaussian RNG functor.
+  /// Construction of uniform RNG functor.
   TensorFillRandomUniformFunc(
     TensorView view_ = TensorView(),
     RandomUniformFunc<Element> func_ = RandomUniformFunc<Element>()
@@ -649,7 +717,7 @@ struct TensorFillRandomUniformFunc {
   }
 };
 
-/// Computes a random Gaussian distribution
+/// Fills the upper or lower part of a symmetric rank-2 tensor with random values of a uniform distribution.
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -669,7 +737,7 @@ struct TensorFillSymmetricRandomUniformFunc {
   // Methods
   //
 
-  /// Construction of Gaussian RNG functor.
+  /// Construction of uniform RNG functor.
   TensorFillSymmetricRandomUniformFunc(
     TensorView view_ = TensorView(),
     RandomUniformFunc<Element> func_ = RandomUniformFunc<Element>(),
@@ -715,7 +783,7 @@ struct TensorFillPadDiagonalRandomUniformFunc {
   // Methods
   //
 
-  /// Construction of Gaussian RNG functor.
+  /// Construction of uniform RNG functor.
   TensorFillPadDiagonalRandomUniformFunc(
     TensorView view_ = TensorView(),
     RandomUniformFunc<Element> func_ = RandomUniformFunc<Element>(),
@@ -747,7 +815,7 @@ struct TensorFillPadDiagonalRandomUniformFunc {
 
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
-/// Fills a tensor with random values with a uniform random distribution.
+/// Fills a tensor with random values of a uniform random distribution.
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -772,7 +840,7 @@ void TensorFillRandomUniform(
   );
 }
 
-/// Fills a tensor with random values with a uniform random distribution.
+/// Fills a tensor with random values of a uniform random distribution.
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -781,10 +849,10 @@ void TensorFillRandomUniform(
   uint64_t seed,                                       ///< seed for RNG
   double max = 1,                                      ///< upper bound of distribution
   double min = 0,                                      ///< lower bound for distribution
-  int bits = -1) {                                     ///< If non-negative, specifies number of fractional bits that 
+  int bits = -1) {                                     ///< If non-negative, specifies number of fractional bits that
                                                        ///  are not truncated to zero. Permits reducing precision of
-                                                       ///  data.                 
-  
+                                                       ///  data.
+
   TensorFillRandomUniform(dst.view_real(), seed, max, min, bits);
   TensorFillRandomUniform(dst.view_imag(), ~seed, max, min, bits);
 }
@@ -879,6 +947,20 @@ void TensorFillPadDiagonalRandomUniform(
 
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
+/// Fills a tensor with a uniform value
+template <
+  typename Element                        ///< Element type
+>
+void BlockFill(
+  Element *ptr,
+  size_t capacity,
+  Element val
+  ) {                                       
+  for (size_t i = 0; i < capacity; ++i) {
+    ReferenceFactory<Element>::get(ptr, i) = val;
+  }
+}
+
 /// Fills a tensor with random values with a uniform random distribution.
 template <
   typename Element                        ///< Element type
@@ -972,7 +1054,7 @@ void TensorFillDiagonal(
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
-/// Helper to fill a tensor's digonal with 1 and 0 everywhere else.
+/// Helper to fill a tensor's diagonal with 1 and 0 everywhere else.
 template <
   typename Element,               ///< Element type
   typename Layout>                ///< Layout function
@@ -1169,6 +1251,37 @@ void TensorFillSequential(
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
+/// Fills a tensor with random values from a distribution.
+template <
+  typename Element,               ///< Element type
+  typename Layout>                ///< Layout function
+void TensorFillRandom(
+  TensorView<Element, Layout> view,       ///< destination tensor
+  uint64_t seed,
+  Distribution dist) {
+
+  using Real = typename RealType<Element>::Type;
+
+  if (dist.kind == Distribution::Gaussian) {
+    TensorFillRandomGaussian(
+      view,
+      seed,
+      dist.gaussian.mean,
+      dist.gaussian.stddev,
+      dist.int_scale);
+  } else if (dist.kind == Distribution::Uniform) {
+    TensorFillRandomUniform(
+      view,
+      seed,
+      dist.uniform.max,
+      dist.uniform.min,
+      dist.int_scale);
+  }
+}
+
+///////////////////////////////////////////////////////////////////////////////////////////////////
+///////////////////////////////////////////////////////////////////////////////////////////////////
+
 /// Fills a block of data with sequential elements
 template <
   typename Element
@@ -1230,7 +1343,8 @@ void BlockFillRandom(
       seed, 
       dist.gaussian.mean, 
       dist.gaussian.stddev, 
-      dist.int_scale);
+      dist.int_scale,
+      dist.gaussian.pnz);
   }
   else if (dist.kind == Distribution::Uniform) {
     BlockFillRandomUniform<Element>(
@@ -1252,7 +1366,7 @@ template <typename Element>
 struct RandomSparseMetaFunc {
   
   uint64_t seed;
-  double range;
+  int range;
   int MetaSizeInBits;
 
   //
@@ -1282,9 +1396,8 @@ struct RandomSparseMetaFunc {
     Element result = 0x0;
 
     for (int i = 0; i < cutlass::sizeof_bits<Element>::value / 4; ++i) {
-      double rnd = double(std::rand()) / double(RAND_MAX);
-      rnd = range * rnd;
-      Element meta = MetaArray[(int)rnd];
+      int rnd = std::rand() % range;
+      Element meta = MetaArray[rnd];
 
       result = (Element)(result | ((Element)(meta << (i * 4))));
     }
@@ -1373,6 +1486,37 @@ void BlockFillRandomSparseMeta(
   }
 }
 
+///////////////////////////////////////////////////////////////////////////////////////////////////
+///////////////////////////////////////////////////////////////////////////////////////////////////
+
+/// Fills a ell block index matrix with random values with a uniform random distribution.
+template <
+  typename Element,                                ///< Element type
+  typename Layout>                                 ///< Layout function
+void TensorFillRandomEllIdx(
+  TensorView<Element, Layout> dst,                 ///< destination tensor
+  uint64_t seed,                                   ///< seed for RNG
+  int rows, int ell_cols, int cols) {              ///< dimension of the matrix 
+
+  std::srand((unsigned)seed);
+
+  for (int i = 0; i < rows; ++i) {
+    int col_idx = std::rand() % cols;
+   
+    for (int j = 0; j < ell_cols; ++j) {
+      dst.at({i, j}) = col_idx;
+
+      if (col_idx != -1) {
+        if (col_idx == (cols - 1)) {
+          col_idx = -1;
+        } else {
+          col_idx = std::rand() % (cols - col_idx - 1) + col_idx + 1;
+        }
+      }
+    }
+  }
+}
+
 ///////////////////////////////////////////////////////////////////////////////////////////////////
 
 /// Copies a diagonal in from host memory without modifying off-diagonal elements.
